/*
 * mousemap: remapping of mouse events
 *
 * MOUSEMAP is distributed under the GNU GENERAL PUBLIC LICENSE (GPL)
 * Version 2 (June 1991). See the "COPYING" file distributed with this
 * software for more info.
 *
 * Copyright (c) 2001 - 2003 by Michael Schroeder <mdhs@gmx.net>
 */

//#define DBG		1
#define MULTI_EVENTS	1

#include <errno.h> 
#include <signal.h> 
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <X11/Xlib.h>
#include <X11/keysym.h>
#include <X11/extensions/XTest.h>

#define PROG		"mousemap"
#define VERSION		"v0.4"

#define PRESS		True
#define RELEASE		False

#define MAXBUTTON	2
#define MINIMUM		6	/* minimum count of buttons */

#define KEYLEFT		"Left"
#define KEYRIGHT	"Right"


#ifdef DBG
# define DBGMSG(x...)	fprintf(stderr, ##x)
#else
# define DBGMSG(x...) 
#endif


static char *symbol[MAXBUTTON] = {KEYLEFT, KEYRIGHT};
static char pressed[MAXBUTTON];
static Display *disp;

/*
 * generate key event
 */
static void fakeKey(char *key, Bool press)
{
  unsigned int keycode;

  keycode = XKeysymToKeycode(disp, XStringToKeysym(key));
  
  XTestFakeKeyEvent(disp, keycode, press, CurrentTime);
}


static void grabButton(int button)
{
  XGrabButton(disp,
	      button,
	      AnyModifier,
	      DefaultRootWindow(disp),
	      False,
	      ButtonPressMask | ButtonReleaseMask,
	      GrabModeAsync,
	      GrabModeAsync,
	      None,
	      None);
}


static void sighandler(int sig)
{
  fprintf(stderr, "%s: terminating on signal %d\n", PROG, sig);

  XUngrabButton(disp, AnyButton, AnyModifier, DefaultRootWindow(disp));
  XCloseDisplay(disp);

  exit(errno);
}


int main(int argc, char *argv[])
{
  unsigned char buf[80], map[16], *beg, *end;
  int buttons, index, i;
  XEvent event;

  if (daemon(0, 1))	/* prog is a daemon and runs in a new session */
    exit(errno);

  if(!(disp = XOpenDisplay(NULL))) {
    fprintf(stderr, "%s: unable to open display\n", PROG);
    exit(1);
  }

  signal(SIGINT, sighandler);
  signal(SIGTERM, sighandler);
  signal(SIGSEGV, sighandler);		/* Oops */

  buttons = XGetPointerMapping(disp, map, sizeof(map));	/* get button count */

  if (buttons < MINIMUM) {
    fprintf(stderr, "%s: mapping is not required %d\n", PROG, buttons);
    XCloseDisplay(disp);
    exit(0);
  }

  for (i = 0; i < MAXBUTTON && i < argc; i++)
    if (argv[i+1])			/* are there user defined symbols? */
      symbol[i] = argv[i+1];

  if (argc > MAXBUTTON && argv[i+1]) {	/* new mapping table? */

    sprintf(buf, "xmodmap -e \"pointer = %s\"", argv[i+1]);

    if (system(buf)) {				 /* success ? */
      XCloseDisplay(disp);
      exit(1);
    }

    XGetPointerMapping(disp, map, sizeof(map));		/* get new map */
  }

  XAllowEvents(disp, AsyncBoth, CurrentTime);	/* release queued events */
  XAllowEvents(disp, SyncBoth, CurrentTime);

  for (i = MINIMUM; i <= buttons; i++) 
    grabButton(i);			/* claim mouse buttons */

  XSelectInput(disp, XDefaultRootWindow(disp),
	       ButtonPressMask | ButtonReleaseMask);

  fprintf(stderr, "%s %s (PID %d) - mapping table:",
	  PROG, VERSION, getpid());
  for (i = 0; i < buttons; i++) {
    if (map[i] >= MINIMUM)
      fprintf(stderr, " %d=%s", map[i], symbol[map[i]-MINIMUM]);
    else
      fprintf(stderr, " %d", map[i]);
  }
  fprintf(stderr, "\n");

  for (;;) {	/* loop forever */

    XNextEvent(disp, &event);		/* block until button pressed */

    DBGMSG("%s: button event: %d\n", PROG, event.xbutton.button);

    index = event.xbutton.button - MINIMUM;	/* 0 = left, 1 = right */

    switch (event.type) {

      case ButtonPress:
	if (!(index & ~1)) {
	  DBGMSG("%s: button pressed: %d (%s); index: %d\n",
		 PROG, event.xbutton.button, symbol[index], index);
	  pressed[index] = 1;
#ifdef MULTI_EVENTS
          for (beg = end = symbol[index]; *end != '\0'; end++) {

            if (*end == ' ') {				/* Separator ? */
              memcpy(buf, beg, end-beg);
              buf[end-beg] = '\0';
              fakeKey(buf, PRESS);
              beg = end+1;
              DBGMSG("%s: pressed: %s\n", PROG, buf);
            }
         }
         if (*beg) {
           fakeKey(beg, PRESS);
           DBGMSG("%s: pressed: %s\n", PROG, beg);
         }
#else
	  fakeKey(symbol[index], PRESS);
#endif
	}
	else {
	  /* Nothing to do - not our button */
	}
	
	break;

      case ButtonRelease:
	if (!(index & ~1)) {
	  DBGMSG("%s: button released: %d (%s); index: %d\n",
		 PROG, event.xbutton.button, symbol[index], index);
	  pressed[index] = 0;
#ifdef MULTI_EVENTS
          for (beg = end = symbol[index]; *end != '\0'; end++) {

            if (*end == ' ') {			/* Separator ? */
              memcpy(buf, beg, end-beg);
              buf[end-beg] = '\0';
              fakeKey(buf, RELEASE);
              beg = end+1;
              DBGMSG("%s: released: %s\n", PROG, buf);
            }
          }
          if (*beg) {
            fakeKey(beg, RELEASE);
            DBGMSG("%s: released: %s\n", PROG, beg);
          }
#else
	  fakeKey(symbol[index], RELEASE);
#endif
	}
	else {
	  /*
	   * Ugly: if any of our buttons are pressed release them now.
	   * This could happen if a button is remapped to key modifier :-(
	   */
	  for (i = 0; i < MAXBUTTON; i++) {
	    if (pressed[i]) {
	      DBGMSG("%s: button released (fake): %d (%s); index: %d\n",
		     PROG, event.xbutton.button, symbol[i], i);
	      pressed[i] = 0;
#ifdef MULTI_EVENTS
              for (beg = end = symbol[index]; *end != '\0'; end++) {

                if (*end == ' ') {				/* Separator ? */
                  memcpy(buf, beg, end-beg);
                  buf[end-beg] = '\0';
                  fakeKey(buf, RELEASE);
                  beg = end+1;
                  DBGMSG("%s: released: %s\n", PROG, buf);
                }
              }
              if (*beg) {
                fakeKey(beg, RELEASE);
                DBGMSG("%s: released: %s\n", PROG, beg);
              }
#else
              fakeKey(symbol[index], RELEASE);
#endif
	    }
	  }
	}
	break;

      default:
	DBGMSG("%s: button default: %d; index: %d\n",
	       PROG, event.xbutton.button, index);

    }
  }

/* never reached */

  return 0;
}




